"""
Welcome to Youtube Autonomous Shortcodes Module.

This shortcode modules is the one that handles
the shortcodes within a text. First, you need
to register the shortcodes you want to be 
detected. Then, the system will detect them and
extract the attributes and content information,
and leave those shortcodes simplified in the 
text.

After this, those shortcodes are searched
again one by one to obtain the previous and the
next word indexes so we are able to know the
moment in which each shortcode must be applied.

TODO: Explain the process in the code: First,
register shortcodes in X file, then add them in
Y file and create a global shortcodes variable
bla bla bla.
"""


from yta_shortcodes.enums import ShortcodeDuration, SimpleShortcodeStart, BlockShortcodeStart
from yta_audio.voice.transcription.objects import AudioTranscription
from yta_general_utils.programming.validator import PythonValidator
from yta_shortcodes.tag_type import ShortcodeTagType
from typing import Union


class Shortcode:
    """
    Base shortcode class that represent the minimum
    a shortcode can have. This class is pretended 
    to be used as a model to create a custom class.

    This class represent a shortcode once its been
    detected in the code with the Shortcode parser
    based on its corresponding shortcode tag that
    was registered to enable the detection, and
    includes the attributes found, the content (if
    existing) and the indexes of the words that 
    were inmediately before the start and end tag.
    """

    name: str
    """
    The shortcode name that represent the shortcode
    meaning.
    """
    _type: ShortcodeTagType
    """
    The shortcode type, that can be 'simple' or
    'block' scoped.
    """
    _context: any
    """
    TODO: I don't know what this is for
    """
    content: Union[None, str]
    """
    The content of the shortcode, that will be None
    if a simple-scoped shortcode, or a string if
    block-scoped.
    """
    attributes: list[dict]
    """
    The list of attributes found when reading the
    shortcode. These attributes can be single values
    (from args) or key-values (from kwargs).
    """
    start_previous_word_index: Union[int, None]
    """
    The index of the word that is inmediately before
    the start tag of this shortcode. Could be None if
    the shortcode is just at the begining. This index
    is considered within the text empty of shortcodes.
    """
    end_previous_word_index: Union[int, None]
    """
    The index of the word that is inmediately before
    the end tag of this shortcode. It is None if the
    shortcode is a simple one. This index is 
    considered within the text empty of shortcodes.
    """

    def __init__(
        self,
        tag: str,
        type: ShortcodeTagType,
        context: any,
        content: str,
        attributes: list[dict],
        start_previous_word_index: Union[int, None] = None,
        end_previous_word_index: Union[int, None] = None,
    ):
        type = ShortcodeTagType.to_enum(type)

        self.tag = tag
        self.type = type
        self._context = context
        self.content = content
        self.attributes = attributes
        self.start_previous_word_index = start_previous_word_index
        self.end_previous_word_index = end_previous_word_index

    @property
    def is_block_scoped(self):
        """
        Check if this shortcode is a block-scoped one
        [shortcode_tag] ... [/shortcode_tag].
        """
        return self._type == ShortcodeTagType.BLOCK
    
    @property
    def is_simple_scoped(self):
        """
        Check if this shortcode is a simple-scoped one
        [shortcode_tag].
        """
        return self._type == ShortcodeTagType.SIMPLE

class YTAShortcode(Shortcode):
    """
    Custom shortcode class that includes the
    ability to calculate the 'start' and 
    'duration' fields from a given transcription
    of the text in which it has been found. It
    will use the words next to the shortcode and
    their transcription time moment to obtain
    this shortcode 'start' and 'duration' values.
    """

    start: Union[BlockShortcodeStart, SimpleShortcodeStart, float, None] = None
    """
    The start time moment in which the shortcode 
    behaviour must be applied. This needs to be 
    calculated with a transcription of the text
    provided when obtaining this shortcode.
    """
    duration: Union[ShortcodeDuration, float, None] = None
    """
    The time the shortcode behaviour must last. This
    needs to be calculated with a transcription of
    the text provided when obtaining this shortcode.
    """

    @property
    def end(self) -> Union[float, None]:
        """
        The end time moment in which the shortcode
        behaviour must end. This is calculated with the
        'start' and 'duration' parameters if set and
        valid and returned as a float number, or as None
        if not able to calculate.
        """
        if (
            PythonValidator.is_number(self.start) and
            PythonValidator.is_number(self.duration)
        ):
            return self.start + self.duration
        
        return None

    def __init__(
        self,
        tag: str,
        type: ShortcodeTagType,
        context: any,
        content: str,
        attributes: list[dict],
        start_previous_word_index: Union[int, None] = None,
        end_previous_word_index: Union[int, None] = None,
    ):
        super().__init__(tag, type, context, content, attributes, start_previous_word_index, end_previous_word_index)
        self.start = float(attributes.get('start', None)) if attributes.get('start', None) is not None else None
        self.duration = float(attributes.get('duration', None)) if attributes.get('duration', None) is not None else None

    def calculate_start_and_duration(
        self,
        transcription: AudioTranscription
    ):
        """
        Processes this shortcode 'start' and 'duration'
        fields by using the 'transcription' if needed
        (if 'start' and 'duration' fields are not numbers
        manually set by the user in the shortcode when
        written).

        This will consider the current 'start' and
        'duration' strategy and apply them to the words
        related to the shortcode to obtain the real 'start'
        and 'duration' number values.
        """
        # TODO: Here below I have simplified the code but
        # it is commented because if the value is not one
        # in the dict it will fail... but I think there is
        # no possibility of being not one in the dict
        if PythonValidator.is_instance(self.start, [SimpleShortcodeStart, BlockShortcodeStart]):
            if self.is_simple_scoped:
                # self.start = {
                #     ShortcodeStart.BETWEEN_WORDS: (transcription.words[self.start_previous_word_index].start + transcription.words[self.start_previous_word_index + 1].start) / 2
                # }[self.start]

                self.start = {
                    SimpleShortcodeStart.BETWEEN_WORDS: (transcription.words[self.start_previous_word_index].start + transcription.words[self.start_previous_word_index + 1].start) / 2
                }.get(self.start, self.start)

                # if self.start == SimpleShortcodeStart.BETWEEN_WORDS:
                #     self.start = (transcription.words[self.start_previous_word_index].start + transcription.words[self.start_previous_word_index + 1].start) / 2
            else:
                # self.start = {
                #     ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD: transcription.words[self.start_previous_word_index + 1].start,
                #     ShortcodeStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD: (transcription.words[self.start_previous_word_index + 1].start + transcription.words[self.start_previous_word_index + 1].end) / 2,
                #     ShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD: transcription.words[self.start_previous_word_index + 1].end
                # }[self.start]

                self.start = {
                    BlockShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD: transcription.words[self.start_previous_word_index + 1].start,
                    BlockShortcodeStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD: (transcription.words[self.start_previous_word_index + 1].start + transcription.words[self.start_previous_word_index + 1].end) / 2,
                    BlockShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD: transcription.words[self.start_previous_word_index + 1].end
                }.get(self.start, self.start)

                # if self.start == BlockShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD:
                #     self.start = transcription.words[self.start_previous_word_index + 1].start
                # elif self.start == BlockShortcodeStart.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD:
                #     self.start = (transcription.words[self.start_previous_word_index + 1].start + transcription.words[self.start_previous_word_index + 1].end) / 2
                # elif self.start == BlockShortcodeStart.END_OF_FIRST_SHORTCODE_CONTENT_WORD:
                #     self.start = transcription.words[self.start_previous_word_index + 1].end

        if PythonValidator.is_instance(self.duration, ShortcodeDuration):
            if self.type == ShortcodeTagType.SIMPLE:
                # self.duration = {
                #     ShortcodeDuration.FILE_DURATION: FILE_DURATION
                # }[self.duration]

                self.duration = {
                    ShortcodeDuration.FILE_DURATION: ShortcodeDuration.FILE_DURATION.value
                }.get(self.duration, self.duration)

                # if self.duration == ShortcodeDuration.FILE_DURATION:
                #     # This duration must be set when the file is ready, so 
                #     # we use a number value out of limits to flag it
                #     # TODO: Maybe I can keep the enum and detect it later
                #     self.duration = ShortcodeDuration.FILE_DURATION.value
            else:
                # self.duration = {
                #     ShortcodeDuration.SHORTCODE_CONTENT: transcription.words[self.previous_end_word_index].end - transcription.words[self.previous_start_word_index + 1].start
                # }[self.duration]

                self.duration = {
                    ShortcodeDuration.SHORTCODE_CONTENT: transcription.words[self.end_previous_word_index].end - transcription.words[self.end_previous_word_index + 1].start
                }.get(self.duration, self.duration)

                # if self.duration == ShortcodeDuration.SHORTCODE_CONTENT:
                #     self.duration = transcription.words[self.end_previous_word_index].end - transcription.words[self.end_previous_word_index + 1].start


"""
This module contains the ShortcodeTag, which
is the shortcode text between square brackets
that will define a shortcode entity within a
text.

This ShortcodeTag is used to be registered in
the shortcode parser to allow it to detect
these kind of shortcodes.

The shortcode tag has a basic handler method
which is used by the parser to handle the
information to make it work properly, but we
have created a 'custom_handler' with which
you can modify the attributes found or do
whatever you need before storing the shortcode
entity foundin the shortcodes array.

Remember that this shortcode tag will be, once
it's been detected in the text' turned into a
shortcode entity that will be handler later
by the software to do something else. You can
modify the way this entity is stored with this
'custom_handler' method. Creat your own
ShortcodeTag class, inherit from this basic
one and make it yours.
"""
from yta_shortcodes.tag_type import ShortcodeTagType
from yta_shortcodes.shortcode import YTAShortcode
from typing import Union


class ShortcodeTag:
    """
    Class that represent a shortcode tag, which
    can be a simple or a block scoped shortcode
    including only one tag [tag_name] or one
    start and end tag [tag_name] ... [/tag_name].

    This shortcode tag is passed to the shortcode
    parser to register it and look for that tag
    in the text you provide to the parser. This
    will handle all the shortcodes with that
    'tag_name', not only one.
    """

    name: str
    """
    The tag name, which is the word that comes
    inmediately after the opening square bracket
    and represent the shortcode.
    """
    _type: ShortcodeTagType
    """
    The type of shortcode tag that can be simple
    or block scoped.
    """

    @property
    def is_block_scoped(
        self
    ) -> bool:
        """
        Check if this shortcode is a block-scoped one
        [shortcode_tag] ... [/shortcode_tag].
        """
        return self._type == ShortcodeTagType.BLOCK
    
    @property
    def is_simple_scoped(
        self
    ) -> bool:
        """
        Check if this shortcode is a simple-scoped one
        [shortcode_tag].
        """
        return self._type == ShortcodeTagType.SIMPLE

    def __init__(
        self,
        name: str,
        type: ShortcodeTagType
    ):
        type = ShortcodeTagType.to_enum(type)

        self.name = name
        self._type = type

    def str(
        self,
        content: Union[str, None],
        pargs,
        kwargs,
    ) -> str:
        """
        Get the output string according to its type,
        that is the representation of this shortcode
        as it was detected, as a string, including
        parameters and everything.
        """
        content = (
            ""
            if content is None else
            content
        )

        attributes = {}
        attributes.update({parg: None for parg in pargs})
        attributes.update(kwargs)

        # TODO: Maybe we can externalize this code to a
        # utils or something to keep it reusable
        tag = f'[{self.name}'
        for key, value in attributes:
            tag += (
                f' {key}'
                if value is None else
                f' {key}={str(value)}'
            )
        tag += (
            f']{content}[/{self.name}]'
            if self.is_block_scoped else
            f']'
        )

        return tag

    def str_simplified(
        self,
        content: Union[str, None]
    ) -> str:
        """
        Get the output string, but simplified, according
        to its type, that is the representation of this
        shortcode as it was detected, as a string, but
        including not the parameters, only the tags and
        the 'content' (if block-scoped and provided).
        """
        content = (
            ""
            if content is None else
            content
        )

        return (
            f'[{self.name}]'
            if self.is_simple_scoped else
            f'[{self.name}]{content}[/{self.name}]'
        )

    def handler(
        self,
        shortcodes,
        pargs,
        kwargs,
        context,
        **extra_args
        #content: Union[str, None] = None
    ):
        """
        The function that handles the shortcode, fills the provided
        'shortcodes' list with a new shortcode object and all the
        required attributes and values.

        :param shortcodes: A list with the shortcodes that have been
        found previously, so we are able to store the new one.
        :param pargs: The positional arguments found in the shortcode
        tag.
        :param kwargs: The key and value arguments found in the
        shortcode tag.
        :param context: The context (I don't know what this is for...)
        :param **extra_args: Extra arguments we want to pass to this
        function.
        """
        # Fill the attributes
        attributes = {}

        for parg in pargs:
            attributes[parg] = None

        if kwargs:
            for kwarg in kwargs:
                attributes[kwarg] = kwargs[kwarg]

        # We handle the 'content' like this to avoid
        # being an strict parameter
        content = extra_args.get('content', None)

        # Here you can customize it a little bit
        # before storing the final Shortcode entity
        self.custom_handler(shortcodes, attributes, context, content)

        shortcodes.append(YTAShortcode(
            tag = self.name,
            type = self._type,
            context = context,
            content = content,
            attributes = attributes
        ))

        # We will remove all the attributes but keep 
        # the shortcode tags and the content (if
        # existing) to be able to detect it again in
        # the text so we can extract the position for
        # extra purposes
        return self.str_simplified(content)
    
    def custom_handler(
        self,
        shortcodes,
        attributes,
        context,
        content: Union[None, str] = None
    ):
        """
        This part of the code will be called just
        before adding the shortcode found to the
        list so you can do whatever you need with
        the 'attributes' parameter that has been
        build with the received 'pargs' and 'kwargs'
        or the 'context' or 'content'.

        You can replace this method in the subclass
        with anything else you need to do, or can
        keep it as it is to have a basic shortcode.
        """
        pass


from yta_general_utils.programming.enum import YTAEnum as Enum


class ShortcodeTagType(Enum):
    """
    The different type of shortcodes that we can handle
    according to their scopes. It could be simple [tag]
    shortcode or a block-scoped one [tag] ... [/tag].
    """
    
    BLOCK = 'block'
    """
    Shortcode type that is built with a start tag [tag], an
    end tag [/tag] and a content between both of those tags.
    """
    SIMPLE = 'simple'
    """
    Shortcode type that is built with a simple tag [tag].
    Also known as atomic shortcode in the native library.
    """


"""
Here it is the shortcode parser, the code that
registers the different shortcode tags to find
in a given text and the handler methods to
extract the attributes of those shortcodes and
anything else that has been programmed to do
with them.

The parser registers the shortcode tags you
provide, detects those tags, extract the
information of them, minimizes the shortcodes,
gets the positions in which they are placed and,
finally, remove them to clean and leave the text
only with no shortcodes.

Any unregistered shortcode is unaccepted.
"""
from yta_shortcodes.tag import ShortcodeTag
from yta_shortcodes.shortcode import YTAShortcode
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression
from yta_general_utils.text.transformer import fix_unseparated_periods, fix_excesive_blank_spaces, fix_separated_parenthesis, add_missing_spaces_before_and_after_parenthesis, fix_separated_square_brackets, add_missing_spaces_before_and_after_square_brackets, fix_ellipsis
from shortcodes import Parser


class ShortcodeParser:
    """
    This class detects in a text the shortcodes
    you register on it, being able to classify
    the content according to the type and what
    is defined on each shortcode you register.

    _The developer must register manually the
    shortcode parser method for each shortcode_.
    """
    
    _parser: Parser = None
    """
    The shortcode parser, for internal use only.
    """
    # TODO: This can be a subclass of YTAShortcode
    # or Shortcode...
    shortcodes: list[YTAShortcode]
    """
    The list of shortcodes that have been found
    in the last analyzed text.
    """

    def __init__(
        self,
        shortcode_tags: list[ShortcodeTag]
    ):
        """
        The 'shortcode_tags' parameters is the list of
        shortcodes we want to detect.
        """
        if any(
            not PythonValidator.is_subclass(shortcode_tag, ShortcodeTag)
            for shortcode_tag in shortcode_tags
        ):
            raise Exception('At least one of the "shortcodes" provided is not a subclass of ShortcodeTag.')

        self._parser = Parser(start = '[', end = ']', inherit_globals = False)

        # We create a function to register individually each
        # shortcode and it's handler so we don't get the
        # handlers overrided
        def register_shortcode_handler(self, shortcode_tag):
            """
            Register the shortcode handler specifically and
            individually for each shortcode.
            """
            def handler(pargs, kwargs, context, content = None):
                return shortcode_tag.handler(
                    self.shortcodes,
                    pargs,
                    kwargs,
                    context,
                    content = content if shortcode_tag.is_block_scoped else None
                )
            
            # Registra el handler individual para ese shortcode
            self._parser.register(
                handler, 
                shortcode_tag.name,
                f'/{shortcode_tag.name}' if shortcode_tag.is_block_scoped else None
            )

        for shortcode_tag in shortcode_tags:
            register_shortcode_handler(self, shortcode_tag)

    def parse(self, text: str):
        """
        Parse the provided 'text' according to the 
        shortcodes that have been registered.
        """
        if not PythonValidator.is_string(text):
            raise Exception('The provided "text" is not a valid string.')
        
        def sanitize_text(text: str):
            """
            Remove anything unwanted from the given 'text'.
            """
            return fix_excesive_blank_spaces(fix_ellipsis(fix_unseparated_periods(fix_separated_parenthesis(fix_separated_square_brackets(add_missing_spaces_before_and_after_parenthesis(add_missing_spaces_before_and_after_square_brackets(text)))))))
        
        self.text = text
        self.text_sanitized = sanitize_text(text)
        self.text_sanitized_with_simplified_shortcodes = ''
        self.text_sanitized_without_shortcodes = ''
        self.shortcodes = []

        """
        This process involves 2 main steps:
        1. Creating shortcode instances from shortcode
        tags found in text and simplified.
        2. Obtain shortcode positions from simplified
        shortcode tags and clean the text.

        First, we parse the text with the shortcode tags
        we have registered on this parser. When they are
        detected, each shortcode tag handler (that has 
        been registered in the parser to handle its
        corresponding shortcode tag) extracts the 
        attributes (and the content if existing) and 
        appends a new shortcode to the list of shortcodes
        found, with all those attributes. In this step,
        the text is simplified so the shortcode tags are
        reduced to only their names. For example,
        [tag attribute=value] turns into [tag].

        Once we've detected the shortcodes, added them
        and simplified the shortcode tags, we iterate
        over all the shortcodes found to obtain its 
        corresponding simplified shortcode tag in the 
        text, so we can get the previous words positions.
        We update the shortcodes found with those 
        positions and we clear the shortcode tags from
        the text.
        """

        # Parse shortcodes, extract the information
        # and fulfill the 'shortcodes' attribute
        self.text_sanitized_with_simplified_shortcodes = self._parser.parse(self.text_sanitized, context = None)

        """
        After this parse we have 'self.shortcodes' fulfilled
        because the parser has the tags registered, and those
        tags are appending the shortcodes found on the
        shortcodes list that is passed as parameter for each
        handling method
        """
        
        words = self.text_sanitized_with_simplified_shortcodes.split(' ')

        # Now lets look for the simplified shortcode
        # positions
        index = 0
        while index < len(words):
            word = words[index]
            if GeneralRegularExpression.SHORTCODE.parse(word):
                # TODO: Improve this to avoid the ones completed
                for shortcode in self.shortcodes:
                    if '/' in word:
                        # End shortcode
                        if shortcode.tag == word.replace('[', '').replace(']', '').replace('/', '') and shortcode.end_previous_word_index is None:
                            # Is that shortcode
                            shortcode.end_previous_word_index = index - 1
                            break
                    else:
                        # Start shortcode
                        if shortcode.tag == word.replace('[', '').replace(']', '') and shortcode.start_previous_word_index is None:
                            # Is that shortcode
                            shortcode.start_previous_word_index = index - 1
                            break

                del(words[index])
            else:
                index += 1

        self.text_sanitized_without_shortcodes = ' '.join(words)

        return self.text_sanitized_without_shortcodes


from yta_general_utils.programming.enum import YTAEnum as Enum


class ShortcodeDuration(Enum):
    """
    Enum class to represent a dynamic value for 
    the shortcode duration so we can calculate it
    later. This duration is valid for both simple
    and block types because it depends on the tag
    (shortode) that is being used.
    """

    FILE_DURATION = 99999
    """
    Represent the duration of a file. This will be
    replaced by the file duration when the 
    shortcode is related to a file and the file is
    downloaded so we know its real duration.
    """
    SHORTCODE_CONTENT = 99998
    """
    Represent the duration of the shortcode 
    content. This means that the shortcode content
    is measured in time and its duration is 
    calculated to be set as this value.
    """

    @classmethod
    def get_default(cls):
        return cls.SHORTCODE_CONTENT
    
class SimpleShortcodeStart(Enum):
    """
    Enum class to represent a dynamic shortcode
    start value for the simple shortcode type.
    This will let us calculate the value
    dynamically when needed according to this
    value.
    """

    BETWEEN_WORDS = 'between_words'
    """
    The shortcode will start just in the middle
    of two words that are dictated in narration.
    After the end of the first one and before 
    the start of the next one.
    """
    
class BlockShortcodeStart(Enum):
    """
    Enum class to represent a dynamic shortcode
    start value for the block shortcode type.
    This will let us calculate the value
    dynamically when needed according to this
    value.
    """

    START_OF_FIRST_SHORTCODE_CONTENT_WORD = 'start_of_first_shortcode_content_word'
    """
    The shortcode will start when the first word
    of its content starts being dictated.
    """
    MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD = 'middle_of_first_shortcode_content_word'
    """
    The shortcode will start when the first word
    of its content is being dictated, just in the
    middle time between the start and the end of
    its dictation.
    """
    END_OF_FIRST_SHORTCODE_CONTENT_WORD = 'end_of_first_shortcode_content_word'
    """
    The shortcode will start when the first word
    of its content finishes being dictated.
    """

    @classmethod
    def get_default(cls):
        return cls.START_OF_FIRST_SHORTCODE_CONTENT_WORD